#
# convenience module for importing canvas mixin classes
#
from ginga.canvas.CompoundMixin import CompoundMixin  # noqa
from ginga.canvas.CanvasMixin import CanvasMixin  # noqa
from ginga.canvas.DrawingMixin import DrawingMixin  # noqa

# END
