# Autogenerated by Astropy-affiliated package ginga's setup.py on 2019-09-21 03:32:45 UTC
from __future__ import unicode_literals
import datetime

version = "3.0.0"
githash = "23f8d97a9b5490057035509062f7f179957aa126"


major = 3
minor = 0
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 9, 21, 3, 32, 45)
debug = False

astropy_helpers_version = "1.2.dev1156"
