/*
 * CSS for Ginga pg widgets
 */

* {
    box-sizing: border-box;
}


html {
    height: 100%;
}

body {
    height: 100%;
    margin: 0px;
}


.hbox, .vbox {
    box-sizing: border-box;

    margin: 0px;

    display: -webkit-flex;
    display: -ms-flexbox;  /* IE 10 */
    display: -ms-flex;     /* IE 11 */
    display: -moz-flex;
    display: flex;

    justify-content: flex-start;
    overflow: visible;
    align-items: stretch;
    align-content: stretch;
}

.hbox {
    -webkit-flex-flow: row;
    -ms-flex-flow: row;
    -moz-flex-flow: row;
    flex-flow: row;
    width: 100%;
}

.vbox {
    -webkit-flex-flow: column;
    -ms-flex-flow: column;
    -moz-flex-flow: column;
    flex-flow: column;
    height: 100%;
}

div {
    box-sizing: border-box;
    white-space: nowrap;
    overflow: visible;
}

canvas {
    top: 0;
    left: 0;
    right: 0;
    bottom: 0;
    width: 100%;
    height: 100%;
}

.ui-dialog .ui-dialog-content  {
    /* Fix Dialog resizing issue where resize shrinks the content too
     * much.  See https://stackoverflow.com/questions/31584544/jquery-ui-dialog-on-resizing-shrinks-the-content-of-the-dialog
     */
    padding: 3.5px;
}

.ui-dialog .ui-dialog-titlebar  {
    /* Adjust the titlebar so it is equal to the content fix */
    margin: 3.5px;
}

/*END*/
